
/**************************************************************************
 *
 *  $Id: ptpmon.c 1.16 2016/11/09 15:31:13 thomas-b TRASH $
 *  $Name:  $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Implements ptpmon.h
 * -----------------------------------------------------------------------
 *  $Log: ptpmon.c $
 *  Revision 1.16  2016/11/09 15:31:13  thomas-b
 *  Timestamp all incoming sync messages, not only L3
 *  Revision 1.15  2016/11/09 15:05:00  thomas-b
 *  Fixed fetch correct of SO_TIMESTAMPs
 *  Revision 1.14  2016/11/09 13:01:59  thomas-b
 *  Switched from syn1588 C-API call getTimestamp to SO_TIMESTAMP functionality
 *  Revision 1.13  2016/10/26 07:12:36  thomas-b
 *  Fixed message length for L2 packets
 *  Revision 1.12  2016/10/26 06:30:09  thomas-b
 *  Fixed L2 issue (wrong interface index)
 *  Revision 1.12  2016/10/25 09:22:35  thomas-b
 *  Tried to fix L2 issue
 *  Revision 1.11  2016/10/04 12:21:18  thomas-b
 *  *** empty log message ***
 *  Revision 1.10  2016/10/04 12:10:54  thomas-b
 *  Check if msg header has to be reset after each successful select
 *  Revision 1.9  2016/09/12 13:53:36  thomas-b
 *  Include correction field into offset and path delay calculation
 *  Revision 1.8  2016/08/30 07:39:45  thomas-b
 *  Fixed network to host value conversion for ptp_time_interval
 *  Set socket into promiscuous mode
 *  Revision 1.7  2016/05/19 05:57:47  thomas-b
 *  Changed TimeInterval to signed integer value, so offset and delay can become negative
 *  Revision 1.6  2016/05/17 15:04:41  thomas-b
 *  Fixed bug that messages are denied when not requesting MTIE data
 *  Revision 1.5  2016/05/17 14:43:07  thomas-b
 *  Send MTIE_REQUEST if appropriate flag is set
 *  Evaluate response in seperate TLV parts, also evaluate MTIE_RESPONSE
 *  Changed print method to also output MTIE values
 *  Revision 1.4  2016/04/21 07:11:24  thomas-b
 *  Changed structure of monitoring TLV and adapted output
 *  Revision 1.3  2016/02/04 07:28:38  thomas-b
 *  Store last sync timestamp when receiving monitoring TLV
 *  Revision 1.2  2015/11/19 09:16:31  thomas-b
 *  Added and implemented ptpmon_evaluate_measurement function
 *  Fetch timestamps via new API method syn1588_getTimestamp
 *  Revision 1.1  2015/11/18 07:43:26  thomas-b
 *  Initial revision
 *
 **************************************************************************/

#define _PTPMON
 #include <ptpmon.h>
#undef _PTPMON

#include <arpa/inet.h>
#include <errno.h>
#include <linux/errqueue.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <linux/net_tstamp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/timex.h>
#include <sys/types.h>
#include <unistd.h>

#if USE_MBG_TSU
  #include <syn1588_c_api.h>
#endif

#ifndef SIOCSHWTSTAMP
  #define SIOCSHWTSTAMP 0x89b0
#endif

static struct ptpmon_params params;
static const char* ptp_port_state_strs[] = PTP_PORT_STATE_STRS;


int ptpmon_init(char *if_name, PTP_NETWORK_PROTOCOL protocol, uint8_t domain, time_t timeout, uint8_t verbosity)
{
  int flg_enable = 1, flg_tstamp = 0, io_sock = -1, ret = 0;
  struct sockaddr_storage addr_storage;

  #if USE_MBG_TSU
    struct hwtstamp_config cfg_hwtstamp;
    struct ifreq ifr_hwtstamp;
  #endif

  pthread_mutex_init(&params.mutex, NULL);

  memset(&addr_storage, 0, sizeof(addr_storage));
  memset(&params, 0, sizeof(params));

  strncpy(params.interface.ifr_name, if_name, IFNAMSIZ);
  strncpy(params.if_hwaddr.ifr_name, if_name, IFNAMSIZ);

  #if USE_MBG_TSU
    memset(&ifr_hwtstamp, 0, sizeof(ifr_hwtstamp));
    memset(&cfg_hwtstamp, 0, sizeof(cfg_hwtstamp));

    strncpy(ifr_hwtstamp.ifr_name, if_name, IFNAMSIZ);
    ifr_hwtstamp.ifr_data = (char*)&cfg_hwtstamp;

    cfg_hwtstamp.tx_type = HWTSTAMP_TX_ON;
    cfg_hwtstamp.rx_filter = HWTSTAMP_FILTER_PTP_V2_SYNC;
  #endif

  io_sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
  if(io_sock >= 0)
  {
    ioctl(io_sock, SIOCGIFHWADDR, &params.if_hwaddr);
    close(io_sock);
  }
  params.protocol = protocol;

  switch(protocol)
  {
    case PTP_NETWORK_PROTOCOL_IP4:
    {
      struct sockaddr_in *addr = (struct sockaddr_in*)&addr_storage;
      addr->sin_family = AF_INET;
      addr->sin_addr.s_addr = INADDR_ANY;
      addr->sin_port = htons(PTP_UDP_EVENT_PORT);
      params.evt_sock = socket(AF_INET, SOCK_DGRAM, 0);
      break;
    }
    case PTP_NETWORK_PROTOCOL_IP6:
    {
      struct sockaddr_in6 *addr = (struct sockaddr_in6*)&addr_storage;
      addr->sin6_family = AF_INET6;
      memcpy(addr->sin6_addr.s6_addr, in6addr_any.s6_addr, 16);
      addr->sin6_port = htons(PTP_UDP_EVENT_PORT);
      params.evt_sock = socket(AF_INET6, SOCK_DGRAM, 0);
      break;
    }
    case PTP_NETWORK_PROTOCOL_IEEE_802_3:
    {
      params.evt_sock = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_1588));
      break;
    }
    default: break;
  }

  #if USE_MBG_TSU
    flg_tstamp |= SOF_TIMESTAMPING_TX_HARDWARE;
    flg_tstamp |= SOF_TIMESTAMPING_RX_HARDWARE;
    flg_tstamp |= SOF_TIMESTAMPING_RAW_HARDWARE;
  #else
    flg_tstamp |= SOF_TIMESTAMPING_TX_SOFTWARE;
    flg_tstamp |= SOF_TIMESTAMPING_RX_SOFTWARE;
    flg_tstamp |= SOF_TIMESTAMPING_SOFTWARE;
  #endif

  if((params.evt_sock >= 0) &&
     #if USE_MBG_TSU
       (ioctl(params.evt_sock, SIOCSHWTSTAMP, &ifr_hwtstamp) >= 0) &&
       (setsockopt(params.evt_sock, SOL_SOCKET, SO_TIMESTAMP, &flg_enable, sizeof(flg_enable)) == 0) &&
       (setsockopt(params.evt_sock, SOL_SOCKET, SO_TIMESTAMPNS, &flg_enable, sizeof(flg_enable)) == 0) &&
     #endif
     (setsockopt(params.evt_sock, SOL_SOCKET, SO_TIMESTAMPING, &flg_tstamp, sizeof(flg_tstamp)) == 0) &&
     (setsockopt(params.evt_sock, SOL_SOCKET, SO_BINDTODEVICE, (void *)&params.interface, sizeof(params.interface)) == 0) &&
     (setsockopt(params.evt_sock, SOL_SOCKET, SO_REUSEADDR, &flg_enable, sizeof(flg_enable)) == 0))
  {
    if(ioctl(params.evt_sock, SIOCGIFFLAGS, &params.interface) == 0)
    {
      params.interface.ifr_flags |= IFF_PROMISC;
      if(ioctl(params.evt_sock, SIOCSIFFLAGS, &params.interface) == 0)
      {
        if((protocol != PTP_NETWORK_PROTOCOL_IEEE_802_3) &&
           (bind(params.evt_sock, (struct sockaddr*)&addr_storage, sizeof(addr_storage)) == 0))
        {
          switch(protocol)
          {
            case PTP_NETWORK_PROTOCOL_IP4:
            {
              struct sockaddr_in *addr = (struct sockaddr_in*)&addr_storage;
              addr->sin_port = htons(PTP_UDP_GENERAL_PORT);
              params.gen_sock = socket(AF_INET, SOCK_DGRAM, 0);
              break;
            }
            case PTP_NETWORK_PROTOCOL_IP6:
            {
              struct sockaddr_in6 *addr = (struct sockaddr_in6*)&addr_storage;
              addr->sin6_port = htons(PTP_UDP_GENERAL_PORT);
              params.gen_sock = socket(AF_INET6, SOCK_DGRAM, 0);
              break;
            }
            default: break;
          }

          if((params.gen_sock >= 0) &&
             (setsockopt(params.gen_sock, SOL_SOCKET, SO_BINDTODEVICE, (void *)&params.interface, sizeof(params.interface)) == 0) &&
             (setsockopt(params.gen_sock, SOL_SOCKET, SO_REUSEADDR, &flg_enable, sizeof(flg_enable)) == 0) &&
             (bind(params.gen_sock, (struct sockaddr*)&addr_storage, sizeof(addr_storage)) == 0))
            ret = 1;
        }
        else ret = 1;
      }
    }
  }

  params.interface.ifr_ifindex = if_nametoindex(if_name);

  params.domain = domain;
  params.timeout = timeout;
  params.verbosity = verbosity;

  #if USE_MBG_TSU
    if(ret)
    {
      syn1588_init_c_api(0);
      if(!((syn1588_init(0) == ENoErr) && (syn1588_configurePTPv2Timestamper(0, protocol, UNICAST, domain, 0, 0) == ENoErr)))
        ret = 0;
    }
  #endif

  return ret;
}


void ptpmon_exit()
{
  pthread_mutex_lock(&params.mutex);
  if(params.evt_sock >= 0)
    close(params.evt_sock);
  if(params.gen_sock >= 0)
    close(params.gen_sock);

  #if USE_MBG_TSU
    syn1588_close(0);
  #endif

  pthread_mutex_unlock(&params.mutex);
  pthread_mutex_destroy(&params.mutex);
}


struct ptpmon_slave *ptpmon_create_slave(struct ptp_port_address_t *addr)
{
  struct ptpmon_slave* slave = (struct ptpmon_slave*)malloc(sizeof(struct ptpmon_slave));
  if(slave)
  {
    memset(slave, 0, sizeof(struct ptpmon_slave));
    pthread_mutex_init(&slave->mutex, NULL);
    slave->ref_cnt = 1;
    memcpy(&slave->addr, addr, sizeof(struct ptp_port_address_t));
  }
  return slave;
}


void ptpmon_destroy_slave(struct ptpmon_slave *slave)
{
  if(slave)
  {
    pthread_mutex_lock(&slave->mutex);
    while(slave->ref_cnt > 1)
    {
      pthread_mutex_unlock(&slave->mutex);
      usleep(1000);
      pthread_mutex_lock(&slave->mutex);
    }
    pthread_mutex_destroy(&slave->mutex);
    free(slave);
  }
}


int ptpmon_get_egress_tstamp(int sock, PTP_NETWORK_PROTOCOL protocol, uint8_t* pkt, size_t pkt_size, struct ptp_timestamp_t *ts)
{
  if((sock < 0) || !pkt || !pkt_size || !ts)
    return 0;

  int res;
  char data[256];
  struct cmsghdr *cmsg;
  struct msghdr msg;
  struct iovec entry;
  struct sockaddr_in from_addr;
  struct
  {
    struct cmsghdr cm;
    char control[512];
  } control;
  struct timeval timeout;
  struct timespec *spec = NULL;

  fd_set set;

  timeout.tv_sec = 1;
  timeout.tv_usec = 0;

  memset(&msg, 0, sizeof(msg));
  msg.msg_iov = &entry;
  msg.msg_iovlen = 1;
  entry.iov_base = data;
  entry.iov_len = sizeof(data);
  msg.msg_name = (caddr_t)&from_addr;
  msg.msg_namelen = sizeof(from_addr);
  msg.msg_control = &control;
  msg.msg_controllen = sizeof(control);

  while(!((timeout.tv_sec == 0) && (timeout.tv_usec == 0)) && !spec)
  {
    res = 0;
    spec = NULL;

    FD_ZERO(&set);
    FD_SET(sock, &set);

    if(select(sock + 1, &set, 0, 0, &timeout) > 0)
    {
      res = recvmsg(sock, &msg, MSG_ERRQUEUE);

      if((res > 0) &&
         (memcmp(pkt, data + res - pkt_size, pkt_size) == 0))
      {
        for(cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg))
        {
          if((cmsg->cmsg_level == SOL_SOCKET) &&
             (cmsg->cmsg_type == SO_TIMESTAMPING))
          {
            spec = (struct timespec *)CMSG_DATA(cmsg);
            #if USE_MBG_TSU
              spec++;
              spec++;
            #endif
          }
        }
      }
    }
  }

  if(spec)
  {
    ptpmon_seconds_to_ptp_timestamp(ts, (uint64_t)spec->tv_sec);
    ts->nanoseconds = spec->tv_nsec;
    return 1;
  }

  return 0;
}


int ptpmon_send_monitoring_request(struct ptpmon_slave *slave, int mtie_flg)
{
  int i = -2;
  if(slave)
  {
    uint8_t buf[sizeof(struct ethhdr) + sizeof(struct ptp_delay_req_monitoring_message) + sizeof(struct ptp_monitoring_mtie_req_tlv)] = { 0 };
    PTP_NETWORK_PROTOCOL protocol = N_PTP_NETWORK_PROTOCOL;
    struct sockaddr_storage addr_storage;

    int sock = -1;

    memset(&addr_storage, 0, sizeof(addr_storage));

    pthread_mutex_lock(&slave->mutex);
    slave->ref_cnt++;
    slave->mon_flags.req_ok = 0;
    memset(&slave->mon_data, 0, sizeof(slave->mon_data));
    pthread_mutex_unlock(&slave->mutex);

    pthread_mutex_lock(&params.mutex);
    sock = params.evt_sock;
    protocol = (PTP_NETWORK_PROTOCOL)params.protocol;
    pthread_mutex_unlock(&params.mutex);

    if(sock >= 0)
    {
      struct ethhdr *ether_header = NULL;
      struct ptp_delay_req_monitoring_message *req = NULL;
      struct ptp_monitoring_mtie_req_tlv *mtie_tlv = NULL;
      int pos = 0;

      switch(protocol)
      {
        case PTP_NETWORK_PROTOCOL_IP4:
        {
          struct sockaddr_in *addr = (struct sockaddr_in*)&addr_storage;
          addr->sin_family = AF_INET;
          addr->sin_port = htons(PTP_UDP_EVENT_PORT);

          pthread_mutex_lock(&slave->mutex);
          memcpy(&addr->sin_addr.s_addr, slave->addr.address, 4);
          pthread_mutex_unlock(&slave->mutex);
          break;
        }
        case PTP_NETWORK_PROTOCOL_IP6:
        {
          struct sockaddr_in6 *addr = (struct sockaddr_in6*)&addr_storage;
          addr->sin6_family = AF_INET6;
          addr->sin6_port = htons(PTP_UDP_EVENT_PORT);

          pthread_mutex_lock(&slave->mutex);
          memcpy(&addr->sin6_addr.s6_addr, slave->addr.address, 16);
          pthread_mutex_unlock(&slave->mutex);

          pthread_mutex_lock(&params.mutex);
          addr->sin6_scope_id = params.interface.ifr_ifindex;
          pthread_mutex_unlock(&params.mutex);
          break;
        }
        case PTP_NETWORK_PROTOCOL_IEEE_802_3:
        {
          struct sockaddr_ll *addr = (struct sockaddr_ll*)&addr_storage;

          ether_header = (struct ethhdr*)&buf[pos];
          pos += sizeof(struct ethhdr);
          ether_header->h_proto = htons(0x88F7);

          pthread_mutex_lock(&params.mutex);
          addr->sll_ifindex = params.interface.ifr_ifindex;
          pthread_mutex_unlock(&params.mutex);
          break;
        }
        default: break;
      }

      req = (struct ptp_delay_req_monitoring_message*)&buf[pos];
      pos += sizeof(struct ptp_delay_req_monitoring_message);
      if(mtie_flg)
      {
        mtie_tlv = (struct ptp_monitoring_mtie_req_tlv*)&buf[pos];
        pos += sizeof(struct ptp_monitoring_mtie_req_tlv);
      }

      pthread_mutex_lock(&params.mutex);
      if(ether_header)
        memcpy(ether_header->h_source, params.if_hwaddr.ifr_hwaddr.sa_data, ETH_ALEN);
      req->header.source_port_id.clock_id.id[0] = params.if_hwaddr.ifr_hwaddr.sa_data[0];
      req->header.source_port_id.clock_id.id[1] = params.if_hwaddr.ifr_hwaddr.sa_data[1];
      req->header.source_port_id.clock_id.id[2] = params.if_hwaddr.ifr_hwaddr.sa_data[2];
      req->header.source_port_id.clock_id.id[5] = params.if_hwaddr.ifr_hwaddr.sa_data[3];
      req->header.source_port_id.clock_id.id[6] = params.if_hwaddr.ifr_hwaddr.sa_data[4];
      req->header.source_port_id.clock_id.id[7] = params.if_hwaddr.ifr_hwaddr.sa_data[5];
      req->header.domain_number = params.domain;
      pthread_mutex_unlock(&params.mutex);

      req->header.source_port_id.clock_id.id[3] = 0xFF;
      req->header.source_port_id.clock_id.id[4] = 0xFE;
      req->header.source_port_id.port_number = htons(1);

      req->header.message_type = PTP_DELAY_REQ_MESSAGE_TYPE;
      req->header.version = PTP_VERSION;
      if(ether_header)
        req->header.message_length = htons(pos - sizeof(*ether_header));
      else req->header.message_length = htons(pos);
      req->header.flag_field.unicast = 1;
      req->header.control_field = PTP_DELAY_REQ_MESSAGE_TYPE;
      req->header.log_message_interval = PTP_LOG_MESSAGE_INTERVAL_ANY;

      req->tlv.header.type = htons(PTPMON_REQ_TLV_TYPE);

      if(mtie_flg)
        mtie_tlv->header.type = htons(PTPMON_MTIE_REQ_TLV_TYPE);

      pthread_mutex_lock(&slave->mutex);
      if(ether_header)
        memcpy(ether_header->h_dest, slave->addr.address, ETH_ALEN);
      req->header.sequence_id = htons(slave->sequence_id);
      pthread_mutex_unlock(&slave->mutex);

      i = sendto(sock, buf, pos, 0, (struct sockaddr*)&addr_storage, sizeof(addr_storage));

      if(i > 0)
      {
        struct ptp_timestamp_t ts_egress;
        if(!ptpmon_get_egress_tstamp(sock, protocol, buf, pos, &ts_egress))
          i = 0;

        if(i > 0)
        {
          pthread_mutex_lock(&slave->mutex);
          slave->mon_data.s2m_measurement.ts_egress = ts_egress;
          slave->mon_flags.req_ok = 1;
          pthread_mutex_unlock(&slave->mutex);
        }
      }
      else i = -1;
    }

    pthread_mutex_lock(&slave->mutex);
    slave->ref_cnt--;
    pthread_mutex_unlock(&slave->mutex);
  }

  return i;
}


void ptpmon_recv_messages(struct ptpmon_slave *slave, uint16_t port)
{
  PTP_NETWORK_PROTOCOL protocol = N_PTP_NETWORK_PROTOCOL;
  struct sockaddr_storage addr_storage;
  uint64_t us_tout, ns_corr;
  uint16_t sequence_id;
  int sock = -1;

  if(!slave)
    return;

  pthread_mutex_lock(&slave->mutex);
  sequence_id = slave->sequence_id;
  pthread_mutex_unlock(&slave->mutex);

  pthread_mutex_lock(&params.mutex);
  protocol = params.protocol;
  if(port == PTP_UDP_GENERAL_PORT)
    sock = params.gen_sock;
  else sock = params.evt_sock;
  us_tout = (uint64_t)params.timeout * 1000000;
  pthread_mutex_unlock(&params.mutex);

  if(sock >= 0)
  {
    struct msghdr msg;
    struct iovec entry;
    struct
    {
      struct cmsghdr cm;
      char control[512];
    } control;
    struct timeval tout;

    char data[256];
    fd_set set;

    memset(&addr_storage, 0, sizeof(addr_storage));
    memset(&msg, 0, sizeof(msg));

    msg.msg_iov = &entry;
    msg.msg_iovlen = 1;
    entry.iov_base = data;
    entry.iov_len = sizeof(data);
    msg.msg_name = (caddr_t)&addr_storage;
    msg.msg_namelen = sizeof(addr_storage);
    msg.msg_control = &control;
    msg.msg_controllen = sizeof(control);

    pthread_mutex_lock(&slave->mutex);
    while(!(us_tout == 0) && (!slave->mon_flags.sync_ok || !slave->mon_flags.follow_up_ok || !slave->mon_flags.resp_ok))
    {
      pthread_mutex_unlock(&slave->mutex);

      FD_ZERO(&set);
      FD_SET(sock, &set);

      if((tout.tv_sec == 0) && (tout.tv_usec == 0))
      {
        tout.tv_sec = 0;
        if(us_tout > 100000)
        {
          tout.tv_usec = 100000;
          us_tout -= 100000;
        }
        else
        {
          tout.tv_usec = us_tout;
          us_tout = 0;
        }
      }

      if(select(sock + 1, &set, NULL, NULL, &tout) > 0)
      {
        int n = recvmsg(sock, &msg, MSG_DONTWAIT);
        struct ethhdr *ether_header = NULL;

        switch(protocol)
        {
          case PTP_NETWORK_PROTOCOL_IP4:
          {
            struct sockaddr_in* addr_in = (struct sockaddr_in*)&addr_storage;
            pthread_mutex_lock(&slave->mutex);
            if(memcmp(&addr_in->sin_addr.s_addr, slave->addr.address, 4) != 0)
              n = 0;
            pthread_mutex_unlock(&slave->mutex);
            break;
          }
          case PTP_NETWORK_PROTOCOL_IP6:
          {
            struct sockaddr_in6* addr_in6 = (struct sockaddr_in6*)&addr_storage;
            pthread_mutex_lock(&slave->mutex);
            if(memcmp(&addr_in6->sin6_addr.s6_addr, slave->addr.address, 16) != 0)
              n = 0;
            pthread_mutex_unlock(&slave->mutex);
            break;
          }
          case PTP_NETWORK_PROTOCOL_IEEE_802_3:
          {
            ether_header = (struct ethhdr*)&data[0];
            if(ether_header->h_proto != ntohs(0x88F7))
              n = 0;
            pthread_mutex_lock(&slave->mutex);
            if(memcmp(ether_header->h_source, slave->addr.address, ETH_ALEN) != 0)
              n = 0;
            pthread_mutex_unlock(&slave->mutex);
            pthread_mutex_lock(&params.mutex);
            if(memcmp(ether_header->h_dest, params.if_hwaddr.ifr_hwaddr.sa_data, ETH_ALEN) != 0)
              n = 0;
            pthread_mutex_unlock(&params.mutex);
            break;
          }
          default: n = 0; break;
        }

        if(n > 0)
        {
          struct ptp_sync_message* ptp_msg = NULL;
          if(ether_header)
            ptp_msg = (struct ptp_sync_message*)&data[sizeof(struct ethhdr)];
          else ptp_msg = (struct ptp_sync_message*)&data[0];

          switch(ptp_msg->header.message_type)
          {
            case PTP_SYNC_MESSAGE_TYPE:
            {
              if((n >= (int)sizeof(struct ptp_sync_message)) &&
                 (ntohs(ptp_msg->header.sequence_id) == sequence_id))
              {
                struct timespec* spec = NULL;
                struct cmsghdr *cmsg;

                for(cmsg = CMSG_FIRSTHDR(&msg); cmsg && !spec; cmsg = CMSG_NXTHDR(&msg, cmsg))
                {
                  if(cmsg->cmsg_level == SOL_SOCKET)
                  {
                    if(cmsg->cmsg_type == SO_TIMESTAMPING)
                    {
                      spec = (struct timespec *)CMSG_DATA(cmsg);
                      #if USE_MBG_TSU
                        spec++;
                        spec++;
                      #endif
                    }

                    if(spec)
                    {
                      pthread_mutex_lock(&slave->mutex);
                      ptpmon_seconds_to_ptp_timestamp(&slave->mon_data.m2s_measurement.ts_ingress, (uint64_t)spec->tv_sec);
                      slave->mon_data.m2s_measurement.ts_ingress.nanoseconds = spec->tv_nsec;
                      pthread_mutex_unlock(&slave->mutex);
                    }
                  }
                }

                if(!spec)
                  n = 0;

                if(n > 0)
                {
                  pthread_mutex_lock(&slave->mutex);
                  if(!ptp_msg->header.flag_field.two_step)
                  {
                    memcpy(slave->mon_data.m2s_measurement.ts_egress.seconds, ptp_msg->origin_timestamp.seconds, 6);
                    slave->mon_data.m2s_measurement.ts_egress.nanoseconds = ntohl(ptp_msg->origin_timestamp.nanoseconds);
                    slave->mon_flags.follow_up_ok = 1;
                  }
                  ns_corr = ptpmon_ptp_timestamp_to_seconds((struct ptp_timestamp_t*)&slave->mon_data.m2s_measurement.correction);
                  ns_corr += ptpmon_ptp_timestamp_to_seconds((struct ptp_timestamp_t*)&ptp_msg->header.correction);
                  ptpmon_seconds_to_ptp_timestamp((struct ptp_timestamp_t*)&slave->mon_data.m2s_measurement.correction, ns_corr);
                  slave->mon_flags.sync_ok = 1;
                  pthread_mutex_unlock(&slave->mutex);
                }
              }
              break;
            }
            case PTP_FOLLOW_UP_MESSAGE_TYPE:
            {
              if((n >= (int)sizeof(struct ptp_sync_message)) &&
                 (ntohs(ptp_msg->header.sequence_id) == sequence_id))
              {
                pthread_mutex_lock(&slave->mutex);
                memcpy(slave->mon_data.m2s_measurement.ts_egress.seconds, ptp_msg->origin_timestamp.seconds, 6);
                slave->mon_data.m2s_measurement.ts_egress.nanoseconds = ntohl(ptp_msg->origin_timestamp.nanoseconds);
                ns_corr = ptpmon_ptp_timestamp_to_seconds((struct ptp_timestamp_t*)&slave->mon_data.m2s_measurement.correction);
                ns_corr += ptpmon_ptp_timestamp_to_seconds((struct ptp_timestamp_t*)&ptp_msg->header.correction);
                ptpmon_seconds_to_ptp_timestamp((struct ptp_timestamp_t*)&slave->mon_data.m2s_measurement.correction, ns_corr);
                slave->mon_flags.follow_up_ok = 1;
                pthread_mutex_unlock(&slave->mutex);
              }
              break;
            }
            case PTP_DELAY_RESP_MESSAGE_TYPE:
            {
              if((n >= (int)(sizeof(struct ptp_delay_resp_monitoring_message) + sizeof(struct ptp_monitoring_resp_tlv_1) +
                             sizeof(uint16_t) + sizeof(uint16_t) + sizeof(struct ptp_monitoring_resp_tlv_2))) &&
                 (ntohs(ptp_msg->header.sequence_id) == sequence_id))
              {
                struct ptp_delay_resp_monitoring_message *resp = NULL;
                struct ptp_monitoring_resp_tlv_1 *tlv_1 = NULL;
                struct ptp_port_address_t *parent_address = NULL;
                struct ptp_monitoring_resp_tlv_2 *tlv_2 = NULL;
                struct ptp_monitoring_mtie_resp_tlv * mtie_tlv = NULL;
                size_t pos = 0;

                if(ether_header)
                  pos += sizeof(struct ethhdr);

                if(n >= (int)(pos + sizeof(struct ptp_delay_resp_monitoring_message)))
                {
                  resp = (struct ptp_delay_resp_monitoring_message*)&data[pos];
                  resp->receive_timestamp.nanoseconds = ntohl(resp->receive_timestamp.nanoseconds);
                  pos += sizeof(struct ptp_delay_resp_monitoring_message);
                }

                if(n >= (int)(pos + sizeof(struct ptp_monitoring_resp_tlv_1)))
                {
                  tlv_1 = (struct ptp_monitoring_resp_tlv_1*)&data[pos];
                  tlv_1->header.type = ntohs(tlv_1->header.type);
                  tlv_1->header.length = ntohs(tlv_1->header.length);
                  pos += sizeof(struct ptp_monitoring_resp_tlv_1);
                }

                if(n >= (int)(pos + sizeof(struct ptp_dyn_port_address_t)))
                {
                  parent_address = (struct ptp_port_address_t*)&data[pos];
                  parent_address->protocol = ntohs(parent_address->protocol);
                  parent_address->length = ntohs(parent_address->length);
                  pos += sizeof(parent_address->protocol) + sizeof(parent_address->length) + parent_address->length;
                }

                if(n >= (int)(pos + sizeof(struct ptp_monitoring_resp_tlv_2)))
                {
                  tlv_2 = (struct ptp_monitoring_resp_tlv_2*)&data[pos];
                  tlv_2->parent_ds.parent_port_id.port_number = ntohs(tlv_2->parent_ds.parent_port_id.port_number);
                  tlv_2->parent_ds.observed_log_variance = ntohs(tlv_2->parent_ds.observed_log_variance);
                  tlv_2->parent_ds.observed_phase_change_rate = ntohl(tlv_2->parent_ds.observed_phase_change_rate);
                  tlv_2->parent_ds.gm_clock_quality.log_variance = ntohs(tlv_2->parent_ds.gm_clock_quality.log_variance);
                  tlv_2->current_ds.steps_removed = ntohs(tlv_2->current_ds.steps_removed);
                  tlv_2->time_properties_ds.current_utc_offset = ntohs(tlv_2->time_properties_ds.current_utc_offset);
                  tlv_2->last_sync_tstamp.nanoseconds = ntohl(tlv_2->last_sync_tstamp.nanoseconds);
                  pos += sizeof(struct ptp_monitoring_resp_tlv_2);
                }

                if(tlv_1 && parent_address && tlv_2 && (tlv_1->header.type == PTPMON_RESP_TLV_TYPE))
                {
                  pthread_mutex_lock(&slave->mutex);
                  slave->mon_data.port_state = tlv_1->port_state;
                  slave->mon_data.parent_address.protocol = parent_address->protocol;
                  slave->mon_data.parent_address.length = parent_address->length;
                  if(parent_address->length <= sizeof(parent_address->address))
                    memcpy(slave->mon_data.parent_address.address, parent_address->address, parent_address->length);
                  slave->mon_data.parent_ds = tlv_2->parent_ds;
                  slave->mon_data.current_ds = tlv_2->current_ds;
                  slave->mon_data.time_properties_ds = tlv_2->time_properties_ds;
                  ptpmon_seconds_to_ptp_timestamp(&slave->mon_data.last_sync_tstamp, ptpmon_ptp_timestamp_to_seconds(&tlv_2->last_sync_tstamp));
                  slave->mon_data.last_sync_tstamp.nanoseconds = ntohl(tlv_2->last_sync_tstamp.nanoseconds);
                  memcpy(slave->mon_data.s2m_measurement.ts_ingress.seconds, resp->receive_timestamp.seconds, 6);
                  slave->mon_data.s2m_measurement.ts_ingress.nanoseconds = resp->receive_timestamp.nanoseconds;
                  ns_corr = ptpmon_ptp_timestamp_to_seconds((struct ptp_timestamp_t*)&slave->mon_data.s2m_measurement.correction);
                  ns_corr += ptpmon_ptp_timestamp_to_seconds((struct ptp_timestamp_t*)&ptp_msg->header.correction);
                  ptpmon_seconds_to_ptp_timestamp((struct ptp_timestamp_t*)&slave->mon_data.s2m_measurement.correction, ns_corr);
                  slave->mon_flags.resp_ok = 1;
                  pthread_mutex_unlock(&slave->mutex);
                }

                if(n >= (int)(pos + sizeof(struct ptp_monitoring_mtie_resp_tlv)))
                {
                  mtie_tlv = (struct ptp_monitoring_mtie_resp_tlv*)&data[pos];
                  mtie_tlv->header.type = ntohs(mtie_tlv->header.type);
                  mtie_tlv->header.length = ntohs(mtie_tlv->header.length);
                  mtie_tlv->mtie_window_number = ntohs(mtie_tlv->mtie_window_number);
                  mtie_tlv->mtie_window_duration = ntohs(mtie_tlv->mtie_window_duration);
                  pos += sizeof(struct ptp_monitoring_mtie_resp_tlv);
                }

                if(mtie_tlv && (mtie_tlv->header.type == PTPMON_MTIE_RESP_TLV_TYPE))
                {
                  pthread_mutex_lock(&slave->mutex);
                  slave->mon_data.mtie_data.mtie_valid = mtie_tlv->mtie_valid;
                  slave->mon_data.mtie_data.mtie_window_number = mtie_tlv->mtie_window_number;
                  slave->mon_data.mtie_data.mtie_window_duration = mtie_tlv->mtie_window_duration;
                  slave->mon_data.mtie_data.min_offs_from_master = mtie_tlv->min_offs_from_master;
                  slave->mon_data.mtie_data.max_offs_from_master = mtie_tlv->max_offs_from_master;
                  pthread_mutex_unlock(&slave->mutex);
                }
              }
              break;
            }
            default: break;
          }
        }

        memset(&msg, 0, sizeof(msg));
        msg.msg_iov = &entry;
        msg.msg_iovlen = 1;
        entry.iov_base = data;
        entry.iov_len = sizeof(data);
        msg.msg_name = (caddr_t)&addr_storage;
        msg.msg_namelen = sizeof(addr_storage);
        msg.msg_control = &control;
        msg.msg_controllen = sizeof(control);
      }
      pthread_mutex_lock(&slave->mutex);
    }
    pthread_mutex_unlock(&slave->mutex);
  }
}


void *ptpmon_recv_event_messages_t(void *t)
{
  struct ptpmon_slave *slave = (struct ptpmon_slave*)t;

  ptpmon_recv_messages(slave, PTP_UDP_EVENT_PORT);

  pthread_mutex_lock(&slave->mutex);
  slave->ref_cnt--;
  pthread_mutex_unlock(&slave->mutex);

  return 0;
}


void *ptpmon_recv_general_messages_t(void *t)
{
  struct ptpmon_slave *slave = (struct ptpmon_slave*)t;

  ptpmon_recv_messages(slave, PTP_UDP_GENERAL_PORT);

  pthread_mutex_lock(&slave->mutex);
  slave->ref_cnt--;
  pthread_mutex_unlock(&slave->mutex);

  return 0;
}


void *ptpmon_recv_messages_t(void *t)
{
  struct ptpmon_slave *slave = (struct ptpmon_slave*)t;

  ptpmon_recv_messages(slave, 0);

  pthread_mutex_lock(&slave->mutex);
  slave->ref_cnt--;
  pthread_mutex_unlock(&slave->mutex);

  return 0;
}


void ptpmon_start_recv_thread(struct ptpmon_slave *slave, uint16_t port)
{
  pthread_t thread;
  pthread_attr_t attr;

  if(!slave)
    return;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  pthread_mutex_lock(&slave->mutex);
  slave->ref_cnt++;
  pthread_mutex_unlock(&slave->mutex);

  if(port == PTP_UDP_EVENT_PORT)
    pthread_create(&thread, &attr, ptpmon_recv_event_messages_t, slave);
  else if(port == PTP_UDP_GENERAL_PORT)
    pthread_create(&thread, &attr, ptpmon_recv_general_messages_t, slave);
  else
    pthread_create(&thread, &attr, ptpmon_recv_messages_t, slave);
}


void ptpmon_start_listening(struct ptpmon_slave *slave)
{
  PTP_NETWORK_PROTOCOL protocol;

  pthread_mutex_lock(&slave->mutex);
  slave->mon_flags.resp_ok = 0;
  slave->mon_flags.sync_ok = 0;
  slave->mon_flags.follow_up_ok = 0;
  protocol = (PTP_NETWORK_PROTOCOL)slave->addr.protocol;
  pthread_mutex_unlock(&slave->mutex);

  if(protocol == PTP_NETWORK_PROTOCOL_IEEE_802_3)
    ptpmon_start_recv_thread(slave, 0);
  else
  {
    ptpmon_start_recv_thread(slave, PTP_UDP_EVENT_PORT);
    ptpmon_start_recv_thread(slave, PTP_UDP_GENERAL_PORT);
  }
}


void ptpmon_seconds_to_ptp_timestamp(struct ptp_timestamp_t *timestamp, uint64_t seconds)
{
  if(timestamp)
  {
    int i;
    for(i = 0; i < 6; i++)
      timestamp->seconds[i] = seconds >> (40 - (i * 8)) & 0xFF;
  }
}


uint64_t ptpmon_ptp_timestamp_to_seconds(struct ptp_timestamp_t *timestamp)
{
  uint64_t seconds = 0;
  if(timestamp)
  {
    int i;
    for(i = 0; i < 6; i++)
      seconds |= (uint64_t)timestamp->seconds[i] << (40 - (i * 8));
  }
  return seconds;
}


uint64_t ptpmon_ptp_correction_to_nanoseconds(struct ptp_correction_t *correction)
{
  uint64_t nanoseconds = 0;
  if(correction)
  {
    int i;
    for(i = 0; i < 6; i++)
      nanoseconds |= (uint64_t)correction->nanoseconds[i] << (40 - (i * 8));
  }
  return nanoseconds;
}


void ptpmon_i64_to_ptp_time_interval(struct ptp_time_interval_t *time_interval, int64_t ns)
{
  if(time_interval)
  {
    int i;
    for(i = 0; i < 6; i++)
      time_interval->ns[i] = (ns >> (40 - (i * 8))) & 0xFF;
  }
}


int64_t ptpmon_ptp_time_interval_to_i64(struct ptp_time_interval_t *time_interval)
{
  int64_t i64 = 0;
  if(time_interval)
  {
    int i;
    for(i = 0; i < 8; i++)
      i64 |= ((int64_t)time_interval->ns[i] & 0xFF) << (56 - (i * 8));
    i64 = i64 >> 16;
  }
  return i64;
}


int ptpmon_evaluate_measurement(struct ptpmon_slave *slave)
{
  int ret = 0;
  int64_t t1_ns = 0, t2_ns = 0, t3_ns = 0, t4_ns = 0, t1_t2_corr_ns = 0, t3_t4_corr_ns = 0;
  pthread_mutex_lock(&slave->mutex);
  if(slave->mon_flags.sync_ok && slave->mon_flags.follow_up_ok)
  {
    t1_ns = ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_egress) * 1000000000;
    t1_ns += slave->mon_data.m2s_measurement.ts_egress.nanoseconds;
    t2_ns = ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_ingress) * 1000000000;
    t2_ns += slave->mon_data.m2s_measurement.ts_ingress.nanoseconds;
    t1_t2_corr_ns = ptpmon_ptp_correction_to_nanoseconds(&slave->mon_data.m2s_measurement.correction);
    slave->mon_data.ns_m2s_delay = t2_ns - t1_ns - t1_t2_corr_ns;
  }
  if(slave->mon_flags.req_ok && slave->mon_flags.resp_ok)
  {
    t3_ns = ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_egress) * 1000000000;
    t3_ns += slave->mon_data.s2m_measurement.ts_egress.nanoseconds;
    t4_ns = ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_ingress) * 1000000000;
    t4_ns += slave->mon_data.s2m_measurement.ts_ingress.nanoseconds;
    t3_t4_corr_ns = ptpmon_ptp_correction_to_nanoseconds(&slave->mon_data.s2m_measurement.correction);
    slave->mon_data.ns_s2m_delay = t4_ns - t3_ns - t3_t4_corr_ns;
  }
  if(slave->mon_flags.req_ok && slave->mon_flags.resp_ok && slave->mon_flags.sync_ok && slave->mon_flags.follow_up_ok)
  {
    slave->mon_data.ns_mean_path_delay = ((t2_ns - t3_ns) + (t4_ns - t1_ns) - t1_t2_corr_ns - t3_t4_corr_ns) / 2;
    slave->mon_data.ns_offset_from_master = t2_ns - t1_ns - slave->mon_data.ns_mean_path_delay - t1_t2_corr_ns;
    ret = 1;
  }
  pthread_mutex_unlock(&slave->mutex);
  return ret;
}


void ptpmon_print_data(struct ptpmon_slave *slave)
{
  uint8_t verbosity;

  if(!slave)
    return;

  pthread_mutex_lock(&params.mutex);
  verbosity = params.verbosity;
  pthread_mutex_unlock(&params.mutex);

  printf("==================================================\n");

  pthread_mutex_lock(&slave->mutex);
  if(verbosity)
  {
    struct ptp_monitoring_data *data = &slave->mon_data;
    char addr[INET6_ADDRSTRLEN] = { 0 };
    uint8_t first_flag = 1;

    if((data->port_state >= PTP_PORT_STATE_UNINITIALIZED) && (data->port_state < N_PTP_PORT_STATE))
      printf("Port State: %s\n", ptp_port_state_strs[data->port_state]);
    else printf("Port State: Unknown\n");
    switch(data->parent_address.protocol)
    {
      case PTP_NETWORK_PROTOCOL_IP4:
      {
        inet_ntop(AF_INET, data->parent_address.address, addr, sizeof(addr));
        break;
      }
      case PTP_NETWORK_PROTOCOL_IP6:
      {
        inet_ntop(AF_INET6, data->parent_address.address, addr, sizeof(addr));
        break;
      }
      case PTP_NETWORK_PROTOCOL_IEEE_802_3:
      {
        snprintf(addr, 18, "%02X:%02X:%02X:%02X:%02X:%02X", data->parent_address.address[0], data->parent_address.address[1],
                 data->parent_address.address[2], data->parent_address.address[3], data->parent_address.address[4],
                 data->parent_address.address[5]);
        break;
      }
      default: break;
    }
    printf("Parent Address: %s\n", addr);
    printf("Parent Clock ID: %02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X\n", data->parent_ds.parent_port_id.clock_id.id[0],
           data->parent_ds.parent_port_id.clock_id.id[1], data->parent_ds.parent_port_id.clock_id.id[2],
           data->parent_ds.parent_port_id.clock_id.id[3], data->parent_ds.parent_port_id.clock_id.id[4],
           data->parent_ds.parent_port_id.clock_id.id[5], data->parent_ds.parent_port_id.clock_id.id[6],
           data->parent_ds.parent_port_id.clock_id.id[7]);
    printf("Parent Port ID: %u\n", data->parent_ds.parent_port_id.port_number);
    printf("Parent Stats: %s\n", data->parent_ds.flags.parent_stats ? "TRUE" : "FALSE");
    printf("Observed Parent Log Variance: %u\n", data->parent_ds.observed_log_variance);
    printf("Observed Parent Phase Change Rate: %d\n", data->parent_ds.observed_phase_change_rate);
    printf("\n");
    printf("Grandmaster Priority 1: %u\n", data->parent_ds.gm_priority_1);
    printf("Grandmaster Clock Class: %u\n", data->parent_ds.gm_clock_quality.clock_class);
    printf("Grandmaster Clock Accuracy: 0x%02X\n", data->parent_ds.gm_clock_quality.clock_accuracy);
    printf("Grandmaster Clock Variance: %u\n", data->parent_ds.gm_clock_quality.log_variance);
    printf("Grandmaster Priority 2: %u\n", data->parent_ds.gm_priority_2);
    printf("Grandmaster Clock ID: %02X:%02X:%02X:%02X:%02X:%02X:%02X:%02X\n", data->parent_ds.gm_identity.id[0],
           data->parent_ds.gm_identity.id[1], data->parent_ds.gm_identity.id[2], data->parent_ds.gm_identity.id[3],
           data->parent_ds.gm_identity.id[4], data->parent_ds.gm_identity.id[5], data->parent_ds.gm_identity.id[6],
           data->parent_ds.gm_identity.id[7]);
    printf("Steps Removed: %u\n", data->current_ds.steps_removed);
    printf("\n");

    printf("Current UTC Offset: %u\n", data->time_properties_ds.current_utc_offset);
    printf("Flags: ");
    if(data->time_properties_ds.flags.li_61)
    {
      if(!first_flag)
        printf(" | ");
      else first_flag = 0;
      printf("LI_61");
    }
    if(data->time_properties_ds.flags.li_59)
    {
      if(!first_flag)
        printf(" | ");
      else first_flag = 0;
      printf("LI_59");
    }
    if(data->time_properties_ds.flags.utc_reasonable)
    {
      if(!first_flag)
        printf(" | ");
      else first_flag = 0;
      printf("UTC_OFFS_REASONABLE");
    }
    if(data->time_properties_ds.flags.timescale)
    {
      if(!first_flag)
        printf(" | ");
      else first_flag = 0;
      printf("PTP_TIMESCALE");
    }
    if(data->time_properties_ds.flags.time_traceable)
    {
      if(!first_flag)
        printf(" | ");
      else first_flag = 0;
      printf("TIME_TRACEABLE");
    }
    if(data->time_properties_ds.flags.frequency_traceable)
    {
      if(!first_flag)
        printf(" | ");
      else first_flag = 0;
      printf("FREQ_TRACEABLE");
    }
    printf("\n");
    printf("Time Source: 0x%02X\n", data->time_properties_ds.time_source);
    printf("\n");

    #if USE_MBG_TSU
      printf("Offset From Master: %lld ns\n", ptpmon_ptp_time_interval_to_i64(&data->current_ds.offset_from_master));
      printf("Mean Path Delay: %lld ns\n", ptpmon_ptp_time_interval_to_i64(&data->current_ds.mean_path_delay));
    #else
      printf("Offset From Master: %ld ns\n", ptpmon_ptp_time_interval_to_i64(&data->current_ds.offset_from_master));
      printf("Mean Path Delay: %ld ns\n", ptpmon_ptp_time_interval_to_i64(&data->current_ds.mean_path_delay));
    #endif
    printf("\n");

    if(data->mtie_data.mtie_valid)
    {
      printf("MTIE Window Number: %u\n", data->mtie_data.mtie_window_number);
      printf("MTIE Window Duration: %u\n", data->mtie_data.mtie_window_duration);

      #if USE_MBG_TSU
        printf("Min. Offset From Master: %lld ns\n", ptpmon_ptp_time_interval_to_i64(&data->mtie_data.min_offs_from_master));
        printf("Max. Offset From Master: %lld ns\n", ptpmon_ptp_time_interval_to_i64(&data->mtie_data.max_offs_from_master));
      #else
        printf("Min. Offset From Master: %ld ns\n", ptpmon_ptp_time_interval_to_i64(&data->mtie_data.min_offs_from_master));
        printf("Max. Offset From Master: %ld ns\n", ptpmon_ptp_time_interval_to_i64(&data->mtie_data.max_offs_from_master));
      #endif

      printf("\n");
    }

    #if USE_MBG_TSU
      printf("Last Sync Tstamp: %lld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.last_sync_tstamp), slave->mon_data.last_sync_tstamp.nanoseconds);
    #else
      printf("Last Sync Tstamp: %ld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.last_sync_tstamp), slave->mon_data.last_sync_tstamp.nanoseconds);
    #endif
    printf("\n");

    #if USE_MBG_TSU
      printf("S2M TX Tstamp: %lld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_egress), slave->mon_data.s2m_measurement.ts_egress.nanoseconds);
      printf("S2M RX Tstamp: %lld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_ingress), slave->mon_data.s2m_measurement.ts_ingress.nanoseconds);
      printf("M2S TX Tstamp: %lld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_egress), slave->mon_data.m2s_measurement.ts_egress.nanoseconds);
      printf("M2S RX Tstamp: %lld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_ingress), slave->mon_data.m2s_measurement.ts_ingress.nanoseconds);
    #else
      printf("S2M TX Tstamp: %ld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_egress), slave->mon_data.s2m_measurement.ts_egress.nanoseconds);
      printf("S2M RX Tstamp: %ld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_ingress), slave->mon_data.s2m_measurement.ts_ingress.nanoseconds);
      printf("M2S TX Tstamp: %ld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_egress), slave->mon_data.m2s_measurement.ts_egress.nanoseconds);
      printf("M2S RX Tstamp: %ld.%d\n", ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_ingress), slave->mon_data.m2s_measurement.ts_ingress.nanoseconds);
    #endif
    printf("\n");
  }

  #if USE_MBG_TSU
    printf("S2M Delay: %lld ns\n", slave->mon_data.ns_s2m_delay);
    printf("M2S Delay: %lld ns\n", slave->mon_data.ns_m2s_delay);
    printf("Mean Path Delay: %lld ns\n", slave->mon_data.ns_mean_path_delay);
    printf("Offset: %lld ns\n", slave->mon_data.ns_offset_from_master);
  #else
    printf("S2M Delay: %ld ns\n", slave->mon_data.ns_s2m_delay);
    printf("M2S Delay: %ld ns\n", slave->mon_data.ns_m2s_delay);
    printf("Mean Path Delay: %ld ns\n", slave->mon_data.ns_mean_path_delay);
    printf("Offset: %ld ns\n", slave->mon_data.ns_offset_from_master);
  #endif

  printf("==================================================\n\n");

  pthread_mutex_unlock(&slave->mutex);
}
