
/**************************************************************************
 *
 *  $Id: ptpmon.h 1.6 2016/05/19 05:57:49 thomas-b TRASH $
 *  $Name:  $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Provides definitions and functions to monitor PTP slave devices
 * -----------------------------------------------------------------------
 *  $Log: ptpmon.h $
 *  Revision 1.6  2016/05/19 05:57:49  thomas-b
 *  Changed TimeInterval to signed integer value, so offset and delay can become negative
 *  Revision 1.5  2016/05/17 14:41:05  thomas-b
 *  Added new TLV types for MTIE_REQUEST and MTIE_RESPONSE
 *  Defined dynamic PORT_ADDRESS structure and split up MONITORING_TLV
 *  Added flag for MTIE_REQUEST to send function
 *  Revision 1.4  2016/04/21 07:11:24  thomas-b
 *  Changed structure of monitoring TLV and adapted output
 *  Revision 1.3  2016/02/04 07:28:42  thomas-b
 *  Added last sync timestamp to ptp_monitoring_resp_tlv
 *  Revision 1.2  2015/11/19 09:16:31  thomas-b
 *  Added and implemented ptpmon_evaluate_measurement function
 *  Fetch timestamps via new API method syn1588_getTimestamp
 *  Revision 1.1  2015/11/18 07:43:26  thomas-b
 *  Initial revision
 *
 **************************************************************************/

#ifndef _PTPMON_H
#define _PTPMON_H

/* Other headers to be included */
#include <gpsdefs.h>

#include <net/if.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sys/time.h>

#ifdef _PTPMON
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif

#define PTP_VERSION                     2

#define PTP_UDP_EVENT_PORT              319
#define PTP_UDP_GENERAL_PORT            320

#define PTPMON_REQ_TLV_TYPE             0x21FE
#define PTPMON_RESP_TLV_TYPE            0x21FF
#define PTPMON_MTIE_REQ_TLV_TYPE        0x2200
#define PTPMON_MTIE_RESP_TLV_TYPE       0x2201

#define PTP_SYNC_MESSAGE_TYPE           0x0
#define PTP_DELAY_REQ_MESSAGE_TYPE      0x1
#define PTP_FOLLOW_UP_MESSAGE_TYPE      0x8
#define PTP_DELAY_RESP_MESSAGE_TYPE     0x9

#define PTP_LOG_MESSAGE_INTERVAL_ANY    0x7F


typedef enum
{
  PTP_NETWORK_PROTOCOL_RESERVED = 0,
  PTP_NETWORK_PROTOCOL_IP4,
  PTP_NETWORK_PROTOCOL_IP6,
  PTP_NETWORK_PROTOCOL_IEEE_802_3,
  PTP_NETWORK_PROTOCOL_DEVICE_NET,
  PTP_NETWORK_PROTOCOL_CONTROL_NET,
  PTP_NETWORK_PROTOCOL_PROFI_NET,
  N_PTP_NETWORK_PROTOCOL
} PTP_NETWORK_PROTOCOL;


#pragma pack(1)

/**
 * @struct ptp_time_interval_t
 * @brief The TimeInterval type represents time intervals (IEEE1588-2008 5.3.2).
 */
struct ptp_time_interval_t
{
  int8_t ns[8];
};


/**
 * @struct ptp_timestamp_t
 * @brief The Timestamp type represents a positive time with respect to the epoch (IEEE1588-2008 5.3.3).
 */
struct ptp_timestamp_t
{
  uint8_t seconds[6];
  uint32_t nanoseconds;
};


/**
 * @struct ptp_correction_t
 * @brief The value of the correction measured in nanoseconds and multiplied by 2^16 (IEEE1588-2008 13.3.2.7).
 */
struct ptp_correction_t
{
  uint8_t nanoseconds[6];
  uint16_t sub_nanoseconds;
};


/**
 * @struct ptp_clock_id_t
 * @brief The ClockIdentity type identifies a clock (IEEE1588-2008 5.3.4).
 */
struct ptp_clock_id_t
{
  uint8_t id[8];
};


/**
 * @struct ptp_port_id_t
 * @brief The PortIdentity type identifies a PTP port (IEEE1588-2008 5.3.5).
 */
struct ptp_port_id_t
{
  struct ptp_clock_id_t clock_id;
  uint16_t port_number;
};


/**
 * @struct ptp_dyn_port_address_t
 * @brief Structure representing a dynamically allocated PTP port address depending on the length field, defined in IEEE 1588-2008
 */
struct ptp_dyn_port_address_t
{
  uint16_t protocol;
  uint16_t length;
  uint8_t *address;
};


/**
 * @struct ptp_port_address_t
 * @brief Structure representing a PTP port address, defined in IEEE 1588-2008
 */
struct ptp_port_address_t
{
  uint16_t protocol;
  uint16_t length;
  uint8_t address[16];
};


/**
 * @struct ptp_clock_quality_t
 * @brief The ClockQuality represents the quality of a clock (IEEE1588-2008 5.3.7).
 */
struct ptp_clock_quality_t
{
  uint8_t clock_class;
  uint8_t clock_accuracy;
  uint16_t log_variance;
};


/**
 * @struct ptp_tlv_header
 * @brief The TLV type represents TLV extension fields (IEEE1588-2008 5.3.8).
 */
struct ptp_tlv_header
{
  uint16_t type;
  uint16_t length;
};


/**
 * @struct ptp_parent_ds_flags
 * @brief Flags for the ::ptp_parent_ds
 */
struct ptp_parent_ds_flags
{
  uint8_t parent_stats    : 1;
  uint8_t reserved        : 7;
};


/**
 * @struct ptp_parent_ds
 * @brief parentDS data set member specifications (IEEE1588-2008 8.2.3).
 */
struct ptp_parent_ds
{
  struct ptp_port_id_t parent_port_id;
  struct ptp_parent_ds_flags flags;
  uint8_t reserved;
  uint16_t observed_log_variance;
  int32_t observed_phase_change_rate;
  uint8_t gm_priority_1;
  struct ptp_clock_quality_t gm_clock_quality;
  uint8_t gm_priority_2;
  struct ptp_clock_id_t gm_identity;
};


/**
 * @struct ptp_current_ds
 * @brief currentDS data set member specifications (IEEE1588-2008 8.2.2).
 */
struct ptp_current_ds
{
  uint16_t steps_removed;
  struct ptp_time_interval_t offset_from_master;
  struct ptp_time_interval_t mean_path_delay;
};


/**
 * @struct ptp_time_properties_flags
 * @brief flags field for timePropertiesDS
 * @see ::ptp_time_properties_ds
 */
struct ptp_time_properties_flags
{
  uint8_t li_61                         : 1;
  uint8_t li_59                         : 1;
  uint8_t utc_reasonable                : 1;
  uint8_t timescale                     : 1;
  uint8_t time_traceable                : 1;
  uint8_t frequency_traceable           : 1;
  uint8_t reserved_1                    : 1;
  uint8_t reserved_2                    : 1;
};


/**
 * @struct ptp_time_properties_ds
 * @brief timePropertiesDS data set member specifications (IEEE1588-2008 8.2.4, 15.5.3.6.1)
 */
struct ptp_time_properties_ds
{
  uint16_t current_utc_offset;
  struct ptp_time_properties_flags flags;
  uint8_t time_source;
};


/**
 * @struct ptp_flag_field
 * @brief The value of the bits of the flagField array shall be as defined in Table 20 (IEEE1588-2008 13.3.2.6).
 */
struct ptp_flag_field
{
  uint8_t alternate_master              : 1;
  uint8_t two_step                      : 1;
  uint8_t unicast                       : 1;
  uint8_t reserved_1                    : 1;
  uint8_t reserved_2                    : 1;
  uint8_t profile_spec_1                : 1;
  uint8_t profile_spec_2                : 1;
  uint8_t security                      : 1;
  uint8_t li_61                         : 1;
  uint8_t li_59                         : 1;
  uint8_t utc_reasonable                : 1;
  uint8_t timescale                     : 1;
  uint8_t time_traceable                : 1;
  uint8_t frequency_traceable           : 1;
  uint8_t reserved_3                    : 1;
  uint8_t reserved_4                    : 1;
};


/**
 * @struct ptp_message_header
 * @brief Structure for the default PTP header used in every PTP message type
 */
struct ptp_message_header
{
  uint8_t message_type                  : 4;
  uint8_t transport_specific            : 4;
  uint8_t version                       : 4;
  uint8_t reserved_a                    : 4;
  uint16_t message_length;
  uint8_t domain_number;
  uint8_t reserved_b;
  struct ptp_flag_field flag_field;
  struct ptp_correction_t correction;
  uint32_t reserved_c;
  struct ptp_port_id_t source_port_id;
  uint16_t sequence_id;
  uint8_t control_field;
  int8_t log_message_interval;
};


/**
 * @struct ptp_monitoring_req_tlv
 * @brief TLV that shall be attached to a delayRequest with monitoring feature.
 */
struct ptp_monitoring_req_tlv
{
  struct ptp_tlv_header header;
};


/**
 * @struct ptp_monitoring_mtie_req_tlv
 * @brief TLV that shall be attached to a delayRequest with monitoring feature to request MTIE statistics.
 */
struct ptp_monitoring_mtie_req_tlv
{
  struct ptp_tlv_header header;
};


/**
 * @struct ptp_monitoring_resp_tlv_1
 * @brief Part 1 of the TLV that shall be attached to a delayResponse in case of a monitoring request.
 */
struct ptp_monitoring_resp_tlv_1
{
  struct ptp_tlv_header header;
  uint8_t port_state;
  uint8_t reserved;
};


/**
 * @struct ptp_monitoring_resp_tlv_2
 * @brief Part 2 of the TLV that shall be attached to a delayResponse in case of a monitoring request.
 */
struct ptp_monitoring_resp_tlv_2
{
  struct ptp_parent_ds parent_ds;
  struct ptp_current_ds current_ds;
  struct ptp_time_properties_ds time_properties_ds;
  struct ptp_timestamp_t last_sync_tstamp;
};


struct ptp_monitoring_mtie_resp_tlv
{
  struct ptp_tlv_header header;
  uint8_t mtie_valid;
  uint8_t reserved;
  uint16_t mtie_window_number;
  uint16_t mtie_window_duration;
  struct ptp_time_interval_t min_offs_from_master;
  struct ptp_time_interval_t max_offs_from_master;
};


/**
 * @struct ptp_sync_message
 * @brief The fields of Sync messages shall be as specified in Table 26 (IEEE1588-2008 13.6.1).
 */
struct ptp_sync_message
{
  struct ptp_message_header header;
  struct ptp_timestamp_t origin_timestamp;
};


/**
 * @struct ptp_delay_req_monitoring_message
 * @brief The fields of Delay Req messages shall be as specified in Table 26 (IEEE1588-2008 13.6.1) + monitoring TLV.
 */
struct ptp_delay_req_monitoring_message
{
  struct ptp_message_header header;
  struct ptp_timestamp_t origin_timestamp;
  struct ptp_monitoring_req_tlv tlv;
};


/**
 * @struct ptp_ip4_delay_resp_monitoring_message
 * @brief The fields of the Delay_Resp message shall be as specified in Table 28 (IEEE1588-2008 13.8.1) + monitoring TLV.
 */
struct ptp_delay_resp_monitoring_message
{
  struct ptp_message_header header;
  struct ptp_timestamp_t receive_timestamp;
  struct ptp_port_id_t requesting_port_id;
};


struct ptp_delay_measurement
{
  struct ptp_timestamp_t ts_egress;
  struct ptp_timestamp_t ts_ingress;
  struct ptp_correction_t correction;
};


/**
 * @struct ptp_mtie_data
 * @brief Structure containing the PTP MTIE data, which has been received in the MTIE TLV
 */
struct ptp_mtie_data
{
  uint8_t mtie_valid;
  uint16_t mtie_window_number;
  uint16_t mtie_window_duration;
  struct ptp_time_interval_t min_offs_from_master;
  struct ptp_time_interval_t max_offs_from_master;
};


/**
 * @struct ptp_monitoring_data
 * @brief Structure containing the PTP monitoring data, which has been calculated or received in the monitoring TLV
 */
struct ptp_monitoring_data
{
  uint8_t port_state;
  struct ptp_port_address_t parent_address;
  struct ptp_parent_ds parent_ds;
  struct ptp_current_ds current_ds;
  struct ptp_time_properties_ds time_properties_ds;
  struct ptp_timestamp_t last_sync_tstamp;
  struct ptp_delay_measurement s2m_measurement;
  struct ptp_delay_measurement m2s_measurement;
  int64_t ns_s2m_delay;
  int64_t ns_m2s_delay;
  int64_t ns_mean_path_delay;
  int64_t ns_offset_from_master;
  struct ptp_mtie_data mtie_data;
};


struct ptp_monitoring_flags
{
  uint8_t req_ok        : 1;
  uint8_t resp_ok       : 1;
  uint8_t sync_ok       : 1;
  uint8_t follow_up_ok  : 1;
  uint8_t reserved      : 4;
};


/**
 * @struct ptpmon_slave
 * @brief Structure containing the needed data for a monitored PTP slave
 */
struct ptpmon_slave
{
  pthread_mutex_t mutex;
  int ref_cnt;
  struct ptp_port_address_t addr;
  uint16_t sequence_id;
  struct ptp_monitoring_flags mon_flags;
  struct ptp_monitoring_data mon_data;
};


struct ptpmon_params
{
  pthread_mutex_t mutex;
  struct ifreq interface;
  struct ifreq if_hwaddr;
  PTP_NETWORK_PROTOCOL protocol;
  int evt_sock;
  int gen_sock;
  uint8_t domain;
  time_t timeout;
  uint8_t verbosity;
};


#pragma pack()

#ifdef __cplusplus
extern "C" {
  #endif

/* Start of header body */

/* function prototypes: */

/* ----- function prototypes begin ----- */

int ptpmon_init(char *if_name, PTP_NETWORK_PROTOCOL protocol, uint8_t domain, time_t timeout, uint8_t verbosity);
void ptpmon_exit();

struct ptpmon_slave *ptpmon_create_slave(struct ptp_port_address_t *addr);
void ptpmon_destroy_slave(struct ptpmon_slave *slave);

int ptpmon_enable_hw_tstamps(int sock);
int ptpmon_get_egress_tstamp(int sock, PTP_NETWORK_PROTOCOL protocol, uint8_t* pkt, size_t pkt_size, struct ptp_timestamp_t *ts);

int ptpmon_send_monitoring_request(struct ptpmon_slave *slave, int mtie_flg);
void ptpmon_start_recv_thread(struct ptpmon_slave *slave, uint16_t port);
void ptpmon_start_listening(struct ptpmon_slave *slave);

void ptpmon_seconds_to_ptp_timestamp(struct ptp_timestamp_t *timestamp, uint64_t seconds);
uint64_t ptpmon_ptp_timestamp_to_seconds(struct ptp_timestamp_t *timestamp);
uint64_t ptpmon_ptp_correction_to_nanoseconds(struct ptp_correction_t *correction);
void ptpmon_i64_to_ptp_time_interval(struct ptp_time_interval_t *time_interval, int64_t ns);
int64_t ptpmon_ptp_time_interval_to_i64(struct ptp_time_interval_t *time_interval);

int ptpmon_evaluate_measurement(struct ptpmon_slave *slave);
void ptpmon_print_data(struct ptpmon_slave *slave);

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

/* (no header definitions found) */

/* ----- function prototypes end ----- */

#ifdef __cplusplus
}
#endif

/* End of header body */

#undef _ext
#undef _DO_INIT

#endif  /* _PTPMON_H */
