
/**************************************************************************
 *
 *  $Id: ptpdate.c 1.5 2016/11/09 15:05:25 thomas-b TRASH $
 *  $Name:  $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Measures S2M and M2S delay for a specific PTP slave device.
 * -----------------------------------------------------------------------
 *  $Log: ptpdate.c $
 *  Revision 1.5  2016/11/09 15:05:25  thomas-b
 *  Calculate UTC offset to PTP timestamps in non TSU mode
 *  Revision 1.4  2016/10/26 06:29:43  thomas-b
 *  Changed usage information
 *  Revision 1.3  2016/05/17 14:41:43  thomas-b
 *  MTIE_REQUEST will now be sent when starting program with '-m'
 *  Revision 1.2  2015/11/19 09:15:55  thomas-b
 *  Added -r parameter option for repeated requests
 *  Revision 1.1  2015/11/18 07:43:26  thomas-b
 *  Initial revision
 *
 **************************************************************************/

#include <ptpmon.h>

#include <arpa/inet.h>
#include <net/if.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void print_usage()
{
  /* show usage information */
  printf("Usage: ptpdate <addr> [-i <str>] [-d <num>] [-t <num>] [-r <num>] [-m] [-v] [-h]\n");
  printf("-d\tPTP domain number  (default: 0)\n");
  printf("-i\tNetwork interface to be used (default: eth1)\n");
  printf("-t\tMaximum time waiting for a response [sec] (default: 2)\n");
  printf("-r\tRepeat request n times (default: 1)\n");
  printf("-m\tAdd request for MTIE statistics\n");
  printf("-v\tVerbose mode\n");
  printf("-h\tPrint usage information\n");
}


int main(int argc, char *argv[])
{
  unsigned int mac_0 = 0, mac_1 = 0, mac_2 = 0, mac_3 = 0, mac_4 = 0, mac_5 = 0;

  struct ptpmon_slave* slave = NULL;
  struct ptp_port_address_t addr;
  char if_name[IFNAMSIZ] = { "eth1" };
  uint8_t domain = 0;
  time_t timeout = 2;
  int verbosity = 0, mtie_flg = 0;
  int i = 1, r = 1, err = 0, j;

  memset(&addr, 0, sizeof(addr));

  if(argc < 2)
  {
    printf("Please specifiy a valid IP address!\n");
    print_usage();
    exit(EXIT_FAILURE);
  }

  while(i < argc)
  {
    if(argv[i][0] == '-')
    {
      switch(argv[i][1])
      {
        case 'h':
        {
          print_usage();
          exit(EXIT_FAILURE);
        }
        case 'd':
        {
          i++;
          domain = (uint8_t)strtoul(argv[i], NULL, 10);
          break;
        }
        case 'i':
        {
          i++;
          strncpy(if_name, argv[i], IFNAMSIZ);
          break;
        }
        case 't':
        {
          i++;
          timeout = (time_t)strtoul(argv[i], NULL, 10);
          if(timeout == 0)
          {
            printf("%s is not a valid timeout!\n", argv[i]);
            print_usage();
            exit(EXIT_FAILURE);
          }
          break;
        }
        case 'r':
        {
          i++;
          r = (uint8_t)strtoul(argv[i], NULL, 10);
          break;
        }
        case 'm':
        {
          mtie_flg = 1;
          break;
        }
        case 'v':
        {
          verbosity = 1;
          break;
        }
        default:
        {
          printf("%c is not a valid parameter!\n", argv[i][1]);
          print_usage();
          exit(EXIT_FAILURE);
        }
      }
    }
    else
    {
      if(inet_pton(AF_INET, argv[i], (struct in_addr*)&addr.address))
      {
        addr.protocol = PTP_NETWORK_PROTOCOL_IP4;
        addr.length = 4;
      }
      else if(inet_pton(AF_INET6, argv[i], (struct in6_addr*)&addr.address))
      {
        addr.protocol = PTP_NETWORK_PROTOCOL_IP6;
        addr.length = 16;
      }
      else if(6 == sscanf(argv[i], "%x:%x:%x:%x:%x:%x", (unsigned int*)&mac_0, (unsigned int*)&mac_1, (unsigned int*)&mac_2,
                                                        (unsigned int*)&mac_3, (unsigned int*)&mac_4, (unsigned int*)&mac_5))
      {
        addr.address[0] = mac_0;
        addr.address[1] = mac_1;
        addr.address[2] = mac_2;
        addr.address[3] = mac_3;
        addr.address[4] = mac_4;
        addr.address[5] = mac_5;
        addr.protocol = PTP_NETWORK_PROTOCOL_IEEE_802_3;
        addr.length = 6;
      }
      else
      {
        printf("%s is not a valid address!\n", argv[i]);
        print_usage();
        exit(EXIT_FAILURE);
      }
    }
    i++;
  }

  if(if_nametoindex(if_name) == 0)
  {
    printf("%s does not exist!\n", if_name);
    print_usage();
    exit(EXIT_FAILURE);
  }

  if(ptpmon_init(if_name, (PTP_NETWORK_PROTOCOL)addr.protocol, domain, timeout, verbosity))
  {
    slave = ptpmon_create_slave(&addr);

    if(slave)
    {
      for(j = 0; j < r; j++)
      {
        err = 0;

        ptpmon_start_listening(slave);
        ptpmon_send_monitoring_request(slave, mtie_flg);

        pthread_mutex_lock(&slave->mutex);
        while(slave->ref_cnt > 1)
        {
          pthread_mutex_unlock(&slave->mutex);
          usleep(1000);
          pthread_mutex_lock(&slave->mutex);
        }

        #if !USE_MBG_TSU
          if(slave->mon_flags.resp_ok && slave->mon_data.time_properties_ds.flags.utc_reasonable)
          {
            if(slave->mon_flags.sync_ok && slave->mon_flags.follow_up_ok)
            {
              ptpmon_seconds_to_ptp_timestamp(&slave->mon_data.m2s_measurement.ts_egress,
                                              ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.m2s_measurement.ts_egress) - slave->mon_data.time_properties_ds.current_utc_offset);
            }

            ptpmon_seconds_to_ptp_timestamp(&slave->mon_data.last_sync_tstamp,
                                            ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.last_sync_tstamp) - slave->mon_data.time_properties_ds.current_utc_offset);

            ptpmon_seconds_to_ptp_timestamp(&slave->mon_data.s2m_measurement.ts_ingress,
                                            ptpmon_ptp_timestamp_to_seconds(&slave->mon_data.s2m_measurement.ts_ingress) - slave->mon_data.time_properties_ds.current_utc_offset);
          }
        #endif

        if(!slave->mon_flags.req_ok)
          printf("Error: Did not send Delay Request!\n");
        if(!slave->mon_flags.resp_ok)
          printf("Error: Did not receive Delay Response!\n");
        if(!slave->mon_flags.sync_ok)
          printf("Error: Did not receive Sync Message!\n");
        if(!slave->mon_flags.follow_up_ok)
          printf("Error: Did not receive Follow Up Message!\n");

        if(!slave->mon_flags.req_ok || !slave->mon_flags.resp_ok ||
           !slave->mon_flags.sync_ok || !slave->mon_flags.follow_up_ok)
          err = 1;
        pthread_mutex_unlock(&slave->mutex);

        if(ptpmon_evaluate_measurement(slave))
          ptpmon_print_data(slave);

        if((j + 1) < r)
          sleep(1);

        pthread_mutex_lock(&slave->mutex);
        slave->sequence_id++;
        pthread_mutex_unlock(&slave->mutex);
      }
      ptpmon_destroy_slave(slave);
    }
    else err = 1;

    ptpmon_exit();
  }
  else err = 1;

  err ? exit(EXIT_FAILURE) : exit(EXIT_SUCCESS);
}
